#!/usr/bin/env python
"""A python module to integrate ITSxpress into QIIME for the trimming of amplicon sequences.

Authors: Adam Rivers and Kyle Weber, USDA Agricultural Research Service

The internally transcribed spacer region is a region between the highly conserved small
subunit (SSU) of rRNA and the large subunit (LSU) of the rRNA. The eukaryotic ITS contains
the 5.8s gene and two variable length spacer regions. In amplicon sequencing studies it is
common practice to trim off the conserved (SSU, 5,8S or LSU) regions. Bengtsson-Palme
et al. (2013) published software the software package ITSx to do this.

ITSxpress is a high-speed implementation of the methods in ITSx than also allows FASTQ
files to be processed. Processing FASTQ files Which is essential for analyzing
sequences using the newer exact Sequence Variant methods in Qiime2, Dada2, Deblur
and Unoise that are replacing OTU clustering.

ITSxpress is also available as a stand-alone software package. See
https://github.com/USDA-ARS-GBRU/itsxpress for details.

"""
import os
import shutil
import math

import yaml
from q2_types.per_sample_sequences import (SingleLanePerSamplePairedEndFastqDirFmt,
                                           SingleLanePerSampleSingleEndFastqDirFmt,
                                           FastqManifestFormat,
                                           YamlFormat)
from q2_types.per_sample_sequences._format import _SingleLanePerSampleFastqDirFmt
from itsxpress import main as itsxpress
from itsxpress.definitions import (taxa_dict,
                   ROOT_DIR)

default_cluster_id=0.987

def _view_artifact_type(per_sample_sequence: _SingleLanePerSampleFastqDirFmt) -> str:
    """Opens the metadata file and looks for the 'type'.
    Args:
        qzaPath: The path of the qza

    Returns:
        (str): The artifact type in the metadata file.

    Raises:
        ValueError: If the metadata file is missing or the 'type' is missing in the metadata file.

    """
    try:
        per_sample_sequence_str = str(per_sample_sequence.path)
        head = os.path.split(per_sample_sequence_str)
        path = os.path.join(str(head[0]), "metadata.yaml")
        with open(path, "r") as fn_open:
            return yaml.safe_load(fn_open)["type"]

    except (NotADirectoryError,
            FileNotFoundError,
            yaml.YAMLError):

        raise ValueError("The metadata file of the qza you entered is missing or the 'type:' in the file is missing.")


def _set_fastqs_and_check(per_sample_sequences: _SingleLanePerSampleFastqDirFmt,
                          artifact_type: str,
                          sequence: dict,
                          single_end: bool,
                          threads: int) -> (str,
                                            object):
    """Checks and writes the fastqs as well as if there paired end, interleaved and single end.

        Args:
            per_sample_sequences (SingleLanePerSampleSingleEndFastqDirFmt): The SingleLanePerSampleSingleEndFastqDirFmt type
            of the input.
            artifact_type (str): The artifact type in the metadata file.
            sequence (dict): The dictionary of sequences and their IDs
            single_end (bool): If the sequences are singled ended or not
            threads (int): The amount of threads to use

        Returns:
            (lists): The sequenceIDs
            (object): The sobj object

        Raises:
            ValueError1: BBtools/ error or fastq format issue.

        """
    # Setting the fastq files and if singleEnd is used.
    fastq = os.path.join(str(per_sample_sequences.path), str(sequence["paths"][0]))
    if "SampleData[PairedEndSequencesWithQuality]" in artifact_type:
        fastq2 = os.path.join(str(per_sample_sequences.path), str(sequence["paths"][1]))
    else:
        fastq2 = None
    sequence_id = sequence["id"]
    # checking fastqs
    try:
        itsxpress._check_fastqs(fastq=fastq, fastq2=fastq2)
        # Parse input types
        paired_end, interleaved = itsxpress._is_paired(fastq=fastq,
                                                       fastq2=fastq2,
                                                       single_end=single_end)
    except (NotADirectoryError,
            FileNotFoundError):

        raise ValueError("There is a problem with the fastq file(s) you selected or\n"
                         "BBtools was not found. check that the BBtools reformat.sh package is executable.")
        # Create SeqSample objects and merge if needed.
    if paired_end and interleaved:
        sobj = itsxpress.SeqSamplePairedInterleaved(fastq=fastq,
                                                    tempdir=None)
        sobj._merge_reads(threads=threads)
        return sequence_id, sobj

    elif paired_end and not interleaved:
        sobj = itsxpress.SeqSamplePairedNotInterleaved(fastq=fastq,
                                                       fastq2=fastq2,
                                                       tempdir=None)
        sobj._merge_reads(threads=threads)
        return sequence_id, sobj

    elif not paired_end and not interleaved:
        sobj = itsxpress.SeqSampleNotPaired(fastq=fastq,
                                            tempdir=None)
        return sequence_id, sobj


def _write_metadata(results: SingleLanePerSampleSingleEndFastqDirFmt):
    """Writes the metadata for the output qza as phred-offset33

    Args:
        results (SingleLanePerSampleSingleEndFastqDirFmt): The SingleLanePerSampleSingleEndFastqDirFmt type
        of the output.

    """

    metadata = YamlFormat()
    metadata.path.write_text(yaml.dump({'phred-offset': 33}))
    results.metadata.write_data(metadata, YamlFormat)


def _fastq_id_maker(per_sample_sequences: _SingleLanePerSampleFastqDirFmt,
                    artifact_type: str) -> (tuple,
                                            bool):
    """Iterates among the manifest to get the file path/name.

    Args:
        per_sample_sequences (SingleLanePerSampleSingleEndFastqDirFmt): The SingleLanePerSampleSingleEndFastqDirFmt type
        of the input.
        artifact_type (str): The artifact type in the metadata file.

    Returns:
        (tuple): The path/name of the sequences.
        (bool): If single end is true or false

    """

    path = os.path.join(str(per_sample_sequences.path), "MANIFEST")
    sequences = []
    with open(path, "r") as fn:
        # {"id": "1", "paths": ["path/to/forward", "path/to/backward"]}
        single_end = False
        for line in fn:
            parts = line.split(",")

            if "#" in line or "sample-id,filename,direction" in line:
                continue

            if not any([sequence['id'] == parts[0] for sequence in sequences]):
                sequence = {"id": parts[0], "paths": [parts[1]]}
                sequences.append(sequence)
                continue

            for sequence in sequences:
                if sequence['id'] == parts[0] and "SampleData[PairedEndSequencesWithQuality]" in artifact_type:
                    sequence["paths"].append(parts[1])

    if (any([len(sequence['paths']) != 2 for sequence in sequences])
            and "SampleData[PairedEndSequencesWithQuality]" in artifact_type):
        raise ValueError("The number of forward and reverse samples do not match.")

    if "SampleData[SequencesWithQuality]" in artifact_type:
        single_end = True
    return sequences, single_end


def _taxa_prefix_to_taxa(taxa_prefix: str) -> str:
    """Turns the taxa prefix letter into the taxa

        Args:
            taxa_prefix (str): The taxa prefix that will be converted to taxa.

        Returns:
            (str): The Taxa

    """
    taxa_dic = {"A": "Alveolata", "B": "Bryophyta", "C": "Bacillariophyta", "D": "Amoebozoa", "E": "Euglenozoa",
                "F": "Fungi", "G": "Chlorophyta", "H": "Rhodophyta", "I": "Phaeophyceae", "L": "Marchantiophyta",
                "M": "Metazoa", "O": "Oomycota", "P": "Haptophyceae", "Q": "Raphidophyceae", "R": "Rhizaria",
                "S": "Synurophyceae", "T": "Tracheophyta", "U": "Eustigmatophyceae", "ALL": "All"}
    taxa_choice = taxa_dic[taxa_prefix]
    return taxa_choice


# First command Trim for SingleLanePerSampleSingleEndFastqDirFmt
def trim_single(per_sample_sequences: SingleLanePerSampleSingleEndFastqDirFmt,
                region: str,
                taxa: str = "F",
                threads: int = 1,
                cluster_id: float = default_cluster_id) -> SingleLanePerSampleSingleEndFastqDirFmt:
    results = main(per_sample_sequences=per_sample_sequences,
                   threads=threads,
                   taxa=taxa,
                   region=region,
                   cluster_id=cluster_id)
    return results


# Second command Trim for SingleLanePerSamplePairedEndFastqDirFmt
def trim_pair(per_sample_sequences: SingleLanePerSamplePairedEndFastqDirFmt,
              region: str,
              taxa: str = "F",
              threads: int = 1,
              cluster_id: float = default_cluster_id) -> SingleLanePerSampleSingleEndFastqDirFmt:
    results = main(per_sample_sequences=per_sample_sequences,
                   threads=threads,
                   taxa=taxa,
                   region=region,
                   cluster_id=cluster_id)
    return results


# The ITSxpress handling
def main(per_sample_sequences: _SingleLanePerSampleFastqDirFmt,
         threads: int,
         taxa: str,
         region: str,
         cluster_id: float) -> SingleLanePerSampleSingleEndFastqDirFmt:
    """The main communication between the plugin and the ITSxpress program.

    Args:
        per_sample_sequences (SingleLanePerSampleSingleEndFastqDirFmt): The SingleLanePerSampleSingleEndFastqDirFmt type
        of the input.
        threads (int) : The number of threads to use.
        taxa (str): The taxa to be used for the search.
        region (str) : The region to be used for the search.
        cluster_id (float):The percent identity for clustering reads, set to 1 for exact dereplication.

    Returns:
        (SingleLanePerSampleSingleEndFastqDirFmt): The SingleLanePerSampleSingleEndFastqDirFmt type
        of the output.

    Raises:
        ValueError1: hmmsearch error.

    """
    #Seeing if cluter_id is equal to 1
    # Finding the artifact type.
    artifact_type = _view_artifact_type(per_sample_sequence=per_sample_sequences)
    # Setting the taxa
    taxa = _taxa_prefix_to_taxa(taxa)
    # Writing the manifest for the output qza
    manifest = FastqManifestFormat()
    manifest_fn = manifest.open()
    manifest_fn.write('sample-id,filename,direction\n')
    # Getting the sequences from the manifest
    sequences, single_end = _fastq_id_maker(per_sample_sequences=per_sample_sequences,
                                            artifact_type=artifact_type)
    barcode = 0
    # Creating result dir
    results = SingleLanePerSampleSingleEndFastqDirFmt()
    # Running the for loop for each sample

    for sequence in sequences:
        # writing fastqs and there attributes and checking the files
        sequence_id, sobj = _set_fastqs_and_check(per_sample_sequences=per_sample_sequences,
                                                  artifact_type=artifact_type,
                                                  sequence=sequence,
                                                  single_end=single_end,
                                                  threads=threads)

        # Deduplicate
        if math.isclose(cluster_id, 1,rel_tol=1e-05):
            sobj.deduplicate(threads=threads)
        else:
            sobj.cluster(threads=threads, cluster_id=cluster_id)
        try:
            # HMMSearch for ITS regions
            hmmfile = os.path.join(ROOT_DIR, "ITSx_db", "HMMs", taxa_dict[taxa])
            sobj._search(hmmfile=hmmfile, threads=threads)
        except (ModuleNotFoundError,
                FileNotFoundError,
                NotADirectoryError):

            raise ValueError("hmmsearch was not found, make sure HMMER3 is installed and executable")

        # Parse HMMseach output.
        its_pos = itsxpress.ItsPosition(domtable=sobj.dom_file,
                                        region=region)
        # Create deduplication object.
        dedup_obj = itsxpress.Dedup(uc_file=sobj.uc_file,
                                    rep_file=sobj.rep_file,
                                    seq_file=sobj.seq_file)

        path_forward = results.sequences.path_maker(sample_id=sequence_id,
                                                    barcode_id=barcode,
                                                    lane_number=1,
                                                    read_number=1)

        manifest_fn.write("{},{},forward\n".format(sequence_id, path_forward.name))
        # Create trimmed sequences.
        dedup_obj.create_trimmed_seqs(str(path_forward),
                                      gzipped=True,
                                      itspos=its_pos)
        # Deleting the temp files.
        shutil.rmtree(sobj.tempdir)
        # Adding one to the barcode
        barcode += 1
    # Writing out the results.
    manifest_fn.close()
    _write_metadata(results=results)
    results.manifest.write_data(manifest,
                                FastqManifestFormat)
    return results



import os
from sys import getsizeof

from nose.tools import eq_, raises
from q2_types.per_sample_sequences import (SingleLanePerSampleSingleEndFastqDirFmt,
                                           SingleLanePerSamplePairedEndFastqDirFmt)

import q2_itsxpress._itsxpress as _itsxpress

# The test data dir
TEST_DIR = os.path.dirname(os.path.abspath(__file__))
# Test info 1
TEST_FILE = os.path.join(TEST_DIR,
                         "test_data",
                         "paired",
                         "445cf54a-bf06-4852-8010-13a60fa1598c",
                         "data")

TEST_DATA = SingleLanePerSamplePairedEndFastqDirFmt(TEST_FILE, "r")
# Test info 2
TEST_FILE_PBMD = os.path.join(TEST_DIR,
                              "test_data",
                              "pairedBrokenMissingData",
                              "50d5f31a-a761-4c04-990c-e7668fe6bf00",
                              "data")

TEST_DATA_PBMD = SingleLanePerSamplePairedEndFastqDirFmt(TEST_FILE_PBMD, "r")
# Test info 3
TEST_FILE_PAF = os.path.join(TEST_DIR,
                             "test_data",
                             "pairedAllForward",
                             "445cf54a-bf06-4852-8010-13a60fa1598c",
                             "data")
TEST_DATA_PAF = SingleLanePerSamplePairedEndFastqDirFmt(TEST_FILE_PAF, "r")
# Test info 4
TEST_FILE_OUT = os.path.join(TEST_DIR,
                             "test_data",
                             "out",
                             "d9955749-00d5-44ae-a628-4b2da43000e1",
                             "data")
TEST_DATA_OUT = SingleLanePerSamplePairedEndFastqDirFmt(TEST_FILE_OUT, "r")
# Test info 5
TEST_FILE_SINGLEOUT = os.path.join(TEST_DIR,
                                   "test_data",
                                   "singleOut",
                                   "75aea4f5-f10e-421e-91d2-feda9fe7b2e1",
                                   "data")
TEST_DATA_SINGLEOUT = SingleLanePerSamplePairedEndFastqDirFmt(TEST_FILE_SINGLEOUT, "r")
# Test info 6
TEST_FILE_SINGLEIN = os.path.join(TEST_DIR,
                                  "test_data",
                                  "singleIn",
                                  "cfd0e65b-05fb-4329-9618-15ecd0aec9b3",
                                  "data")
TEST_DATA_SINGLEIN = SingleLanePerSampleSingleEndFastqDirFmt(TEST_FILE_SINGLEIN, "r")
# Test artifact1
ARTIFACT_TYPE_P = "SampleData[PairedEndSequencesWithQuality]"
# Test artifact2
ARTIFACT_TYPE_S = "SampleData[SequencesWithQuality]"


def test_view_artifcat_type():
    exp1 = _itsxpress._view_artifact_type(per_sample_sequence=TEST_DATA)
    eq_("SampleData[PairedEndSequencesWithQuality]", exp1)
    raises(ValueError, lambda: _itsxpress._view_artifact_type(per_sample_sequence=TEST_DATA_PBMD))


def test_write_metadata():
    results = SingleLanePerSampleSingleEndFastqDirFmt()
    _itsxpress._write_metadata(results)
    path = results.path
    metadata = os.path.join(str(path), "metadata.yml")
    with open(metadata, "rt") as fn:
        eq_("{phred-offset: 33}", fn.readline().replace("\n", ""))


def test_fastq_id_maker():
    exp1, exp2 = _itsxpress._fastq_id_maker(per_sample_sequences=TEST_DATA,
                                            artifact_type=ARTIFACT_TYPE_P)
    for sequence in exp1:
        eq_(sequence["paths"][0], '4774-1-MSITS3_0_L001_R1_001.fastq.gz')
        eq_(sequence["paths"][1], '4774-1-MSITS3_1_L001_R2_001.fastq.gz')
        eq_(exp2, False)
    raises(ValueError, lambda: _itsxpress._fastq_id_maker(per_sample_sequences=TEST_DATA_PAF,
                                                          artifact_type=ARTIFACT_TYPE_P))
    exp3 = _itsxpress._fastq_id_maker(per_sample_sequences=TEST_DATA,
                                      artifact_type=ARTIFACT_TYPE_S)
    eq_(exp3[1], True)


def test_taxa_prefix_to_taxa():
    exp1 = _itsxpress._taxa_prefix_to_taxa(taxa_prefix="A")
    eq_(exp1, "Alveolata")


def test_set_fastqs_and_check():
    sequences, single_end = _itsxpress._fastq_id_maker(per_sample_sequences=TEST_DATA,
                                                       artifact_type=ARTIFACT_TYPE_P)
    for sequence in sequences:
        exp1 = _itsxpress._set_fastqs_and_check(per_sample_sequences=TEST_DATA,
                                                artifact_type=ARTIFACT_TYPE_P,
                                                sequence=sequence,
                                                single_end=single_end,
                                                threads=1)
        eq_(exp1[0], "4774-1-MSITS3")

    sequences2, single_end2 = _itsxpress._fastq_id_maker(per_sample_sequences=TEST_DATA_PAF,
                                                         artifact_type=ARTIFACT_TYPE_S)
    for sequence2 in sequences2:
        exp2 = _itsxpress._set_fastqs_and_check(per_sample_sequences=TEST_DATA,
                                                artifact_type=ARTIFACT_TYPE_S,
                                                sequence=sequence2,
                                                single_end=single_end2,
                                                threads=1)
        eq_(exp2[0], "4774-1-MSITS3")

    sequences3= _itsxpress._fastq_id_maker(per_sample_sequences=TEST_DATA_PAF,
                                           artifact_type=ARTIFACT_TYPE_S)
    for sequence3 in sequences3[0]:
        exp3 = _itsxpress._set_fastqs_and_check(per_sample_sequences=TEST_DATA,
                                                artifact_type=ARTIFACT_TYPE_S,
                                                sequence=sequence3,
                                                single_end=True,
                                                threads=1)
        eq_(exp3[0], "4774-1-MSITS3")


def test_trim_pair():
    threads = 1
    taxa = "F"
    region = "ITS2"

    exp1 = _itsxpress.trim_pair(per_sample_sequences=TEST_DATA,
                                threads=threads,
                                taxa=taxa,
                                region=region)
    exp2 = getsizeof(exp1)
    exp3 = getsizeof(TEST_DATA_OUT)
    eq_(exp2, exp3)


def test_trim_single():
    threads = 1
    taxa = "F"
    region = "ITS2"

    exp1 = _itsxpress.trim_single(per_sample_sequences=TEST_DATA_SINGLEIN,
                                threads=threads,
                                taxa=taxa,
                                region=region)
    exp2 = getsizeof(exp1)
    exp3 = getsizeof(TEST_DATA_SINGLEOUT)
    eq_(exp2, exp3)

def test_trim_single_no_cluster():
    threads = 1
    taxa = "F"
    region = "ITS2"
    cluster_id = 1

    exp1 = _itsxpress.trim_single(per_sample_sequences=TEST_DATA_SINGLEIN,
                                threads=threads,
                                taxa=taxa,
                                region=region,
                                cluster_id=cluster_id)
    exp2 = getsizeof(exp1)
    exp3 = getsizeof(TEST_DATA_SINGLEOUT)
    eq_(exp2, exp3)

def test_trim_pair_no_hmmer():
    threads = 1
    taxa = "F"
    region = "ITS2"

    raises(ValueError, lambda: _itsxpress.trim_pair(per_sample_sequences=TEST_DATA,
                                                    threads=threads,
                                                    taxa=taxa,
                                                    region=region))


def test_trim_pair_no_bb():
    sequences, single_end = _itsxpress._fastq_id_maker(per_sample_sequences=TEST_DATA,
                                                       artifact_type=ARTIFACT_TYPE_P)
    for sequence in sequences:
        raises(ValueError, lambda: _itsxpress._set_fastqs_and_check(per_sample_sequences=TEST_DATA,
                                                                    artifact_type=ARTIFACT_TYPE_P,
                                                                    sequence=sequence,
                                                                    single_end=single_end,
                                                                    threads=1))
        raises(ValueError, lambda: _itsxpress._set_fastqs_and_check(per_sample_sequences=TEST_DATA,
                                                                    artifact_type=ARTIFACT_TYPE_S,
                                                                    sequence=sequence,
                                                                    single_end=single_end,
                                                                    threads=1))



from setuptools import setup

setup(
    name='q2_itsxpress',
    version='1.6.0',
    packages=['q2_itsxpress'],
    author='Adam R. Rivers, Kyle C. Weber',
    author_email='adam.rivers@ars.usda.gov, kweber1@ufl.edu',
    description="A QIIME2 plugin to trim ITS regions using ITSxpress",
    long_description=open('README.rst').read(),
    url='https://github.com/usda-ars-gbru/2_itsxpress',
    test_suite='nose.collector',
    license="License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
    classifiers=["Topic :: Scientific/Engineering :: Bio-Informatics",
                 "Programming Language :: Python :: 3.6",
                 "Programming Language :: Python :: 3.5",
                 "Development Status :: 3 - Alpha"],
    keywords="Amplicon sequencing fungal ITS QIIME2",
    python_requires=">3.5",
    include_package_data=True,
    install_requires=[
        'itsxpress>=1.6.0'
    ],
    entry_points={
        'qiime2.plugins': ['q2_itsxpress=q2_itsxpress.plugin_setup:plugin']
    },
    zip_safe=False
)
